#!/usr/bin/env python
# -*- coding: utf-8 -*-

# |>
# | DATA FORMATS ANALYZER 1-1-1 - DFA PREVIEWER PY [http://www.universalchemicalmarkup.org]
# | Updated: 2015-08-17
# | License: GNU GPL 3 [http://www.gnu.org/licenses/gpl-3.0.html]
# | Author: Jan Mokrý (Main corresponding author:
# |                    Contact information is at the project website.)
# |>

import re

import lxml.etree

import DfaOutput

# DFA PREVIEWER PY - PREVIEW SERVICE
class PreviewService(object):
    '''
    Provides methods to generate the updated preview from XML data files.

    Concepts data are stored in the _`extractedConcepts` *dictionary* with the following structure::

        DICTIONARY{
            Concept name (STRING):DICTIONARY{
                attributes (STRING):LIST[Attribute name (STRING)],
                elements (STRING):LIST[Element name (STRING)],
                types (STRING):LIST[Type name (STRING)]
            }
        }

    Descriptions data are stored in the _`extractedDescriptions` *dictionary* with the following structure::

        DICTIONARY{
            attributes (STRING):DICTIONARY{Attribute name (STRING): Attribute documentation (STRING)},
            elements (STRING):DICTIONARY{Element name (STRING): Element documentation (STRING)},
            types (STRING):DICTIONARY{Type name (STRING): Type documentation (STRING)}
        }

    **Parameters:**

    - verbosity: The *string* specifying verbosity. See details in :ref:`InitService`.
    - filepath: The *string* specifying the full path to XML data files, which will be parsed.
    - version: The *string* specifying the output version. See details in :ref:`InitService`.
    - refName: The *string* specifying the reference name. See details in :ref:`InitService`.
    - refVersion: The *string* specifying the reference version. See details in :ref:`InitService`.
    '''

    def __init__(self, verbosity='IWE', filepath=None, version=None, refName=None, refVersion=None):
        '''
        Initializes class variables.
        '''
        self.outputService = DfaOutput.OutputService(verbosity=verbosity)
        self.refName = refName
        self.refVersion = refVersion
        self.sourceFilenames = [
            filepath + '/Ref--' + self.refName + '--' + self.refVersion + '--Attributes--ExtractedDescription--' + version + '.xml',
            filepath + '/Ref--' + self.refName + '--' + self.refVersion + '--Elements--ExtractedDescription--' + version + '.xml',
            filepath + '/Ref--' + self.refName + '--' + self.refVersion + '--Types--ExtractedDescription--' + version + '.xml']
        self.outputFilenames = [
            filepath + '/Ref--' + self.refName + '--' + self.refVersion + '--AllExtractedInfoAndDescription--' + version + '.xhtml',
            filepath + '/OvRef--MergedConcepts--' + version + '.xml']
        self.extractedConcepts = {}
        self.extractedDescriptions = {}

    def previewData(self):
        '''
        Generates the updated preview from XML data files.
        '''
        self.outputService.printMessage(message='INFO: Loading concepts and descriptions data from XML data files.')
        self._loadData()
        self._updateData()
        self.outputService.printMessage(message='INFO: All concepts and descriptions data updated in the preview.')

    def _loadData(self):
        '''
        Loads data from XML data files.
        '''
        for filepath in self.sourceFilenames:
            try:
                sourceFile = lxml.etree.parse(filepath)
                dataContainer = sourceFile.getroot()
                dataContainerName = dataContainer.tag
                aPattern = re.compile('<att>(.*?)</att>')
                ePattern = re.compile('<el>(.*?)</el>')

                for dataSubContainer in dataContainer.findall(dataContainerName.rstrip('s')):
                    # Load data as unicode objects.
                    name = unicode(dataSubContainer.get('name'), 'utf-8')
                    concept = unicode(dataSubContainer.find('concept').text or '', 'utf-8')
                    description = unicode(lxml.etree.tostring(dataSubContainer.find('description'),
                        encoding='utf-8').strip().replace('<description>\n', '').replace('\n</description>', ''), 'utf-8')
                    # Create links on marked element and attribute names.
                    description = aPattern.sub(r'<a href="#att-\1">\1</a>', description)
                    description = ePattern.sub(r'<a href="#el-\1">\1</a>', description)

                    if concept not in self.extractedConcepts:
                        self.extractedConcepts[concept] = {dataContainerName:[name]}
                    elif dataContainerName not in self.extractedConcepts[concept]:
                        self.extractedConcepts[concept][dataContainerName] = [name]
                    elif name not in self.extractedConcepts[concept][dataContainerName]:
                        self.extractedConcepts[concept][dataContainerName].append(name)

                    if dataContainerName not in self.extractedDescriptions:
                        self.extractedDescriptions[dataContainerName] = {name:description}
                    elif name not in self.extractedDescriptions[dataContainerName]:
                        self.extractedDescriptions[dataContainerName][name] = description
            except IOError:
                self.outputService.printMessage(message='WARNING: File not found - "' + filepath + '"!')

    def _updateData(self):
        '''
        Updates the preview using data loaded from XML data files.
        '''
        self.outputService.printMessage(message='INFO: Updating concepts and descriptions data for XHTML preview.')

        try:
            namespace = '{http://www.w3.org/1999/xhtml}'
            sourceFile = lxml.etree.parse(self.outputFilenames[0])
            contentContainer = sourceFile.getroot().find(namespace + 'body').find(namespace + 'div[@class="content"]')

            if contentContainer.find(namespace + 'div[@class="subcontent concept"]') is not None:
                # Remove old concepts data.
                for divContainer in contentContainer.findall(namespace + 'div[@class="subcontent concept"]'):
                    contentContainer.remove(divContainer)
            else:
                # Create header for concepts section if no previous concepts data were found.
                headingContainer = lxml.etree.Element(namespace + 'h2')
                headingContainer.text = 'Concepts'
                headingContainer.tail = '\n\n'
                contentContainer.append(headingContainer)

            # Add new concepts data.
            for conceptName in sorted(self.extractedConcepts.keys()):
                divContainer = lxml.etree.Element(namespace + 'div')
                divContainer.set('class', 'subcontent concept')
                divContainer.text = '\n'
                divContainer.tail = '\n\n'
                headingContainer = lxml.etree.Element(namespace + 'h3')
                headingContainer.text = conceptName
                headingContainer.tail = '\n'
                divContainer.append(headingContainer)
                listContainer = lxml.etree.Element(namespace + 'ul')
                listContainer.text = '\n'
                listContainer.tail = '\n'
                listItemContainer = lxml.etree.Element(namespace + 'li')

                # The CDATA section is used to ensure that the text containing the listings of attributes, elements and types
                # for the given concept is not escaped, so all markup inside the text will work.
                if 'attributes' in self.extractedConcepts[conceptName]: listItemContainer.text = lxml.etree.CDATA('Attributes: ' +
                    ', '.join('<a href="#att-' + i + '">' + i + '</a>' for i in self.extractedConcepts[conceptName]['attributes']))
                else: listItemContainer.text = 'Attributes: NONE'

                listItemContainer.tail = '\n'
                listContainer.append(listItemContainer)
                listItemContainer = lxml.etree.Element(namespace + 'li')

                if 'elements' in self.extractedConcepts[conceptName]: listItemContainer.text = lxml.etree.CDATA('Elements: ' +
                    ', '.join('<a href="#el-' + i + '">' + i + '</a>' for i in self.extractedConcepts[conceptName]['elements']))
                else: listItemContainer.text = 'Elements: NONE'

                listItemContainer.tail = '\n'
                listContainer.append(listItemContainer)
                listItemContainer = lxml.etree.Element(namespace + 'li')

                if 'types' in self.extractedConcepts[conceptName]: listItemContainer.text = lxml.etree.CDATA('Types: ' +
                    ', '.join('<a href="#t-' + i + '">' + i + '</a>' for i in self.extractedConcepts[conceptName]['types']))
                else: listItemContainer.text = 'Types: NONE'

                listItemContainer.tail = '\n'
                listContainer.append(listItemContainer)
                divContainer.append(listContainer)
                contentContainer.append(divContainer)

            # Update descriptions data.
            for divContainerClass in self.extractedDescriptions.keys():
                for divContainer in contentContainer.findall(namespace
                                                             + 'div[@class="subcontent ' + divContainerClass.rstrip('s') + '"]'):
                    try:
                        divSubContainer = divContainer.find(namespace + 'div[@class="description"]')
                        divSubContainer.clear()
                        divSubContainer.set('class', 'description')
                        # The CDATA section is used to ensure that the text containing the description is not escaped,
                        # so all markup inside the text will work.
                        divSubContainer.text = lxml.etree.CDATA('\n'
                            + self.extractedDescriptions[divContainerClass][divContainer.find(namespace
                                + 'h3[@id]').get('id').split('-')[-1]] + '\n')
                        divSubContainer.tail = '\n'
                    except KeyError:
                        self.outputService.printMessage(message='WARNING: It appears that you removed the item with ID - "'
                            + divContainer.find(namespace + 'h3[@id]').get('id') + '" from this reference!')
                        self.outputService.printMessage(message='INFO: The container of removed item will be deleted now.')
                        self.outputService.printMessage(
                            message='INFO: References like links to the container of removed item must be deleted manually!')
                        contentContainer.remove(divContainer)

            # Write the updated XHTML structure to the output file.
            outputFile = file(self.outputFilenames[0], 'w')
            outputFile.write(lxml.etree.tostring(lxml.etree.ProcessingInstruction(
                'xml', text='version="1.0" encoding="utf-8"'), encoding='utf-8') + '\n')
            # The CDATA sections are removed before output is written to the output file.
            outputFile.write(lxml.etree.tostring(sourceFile, encoding='utf-8').replace(
                '<html', '\n<html').replace('<![CDATA[', '').replace(']]>', ''))
            outputFile.close()
        except IOError:
            self.outputService.printMessage(message='ERROR: File not found - "' + self.outputFilenames[0] + '"!')

        self.outputService.printMessage(message='INFO: Updating concepts data for XML preview.')

        try:
            sourceFile = lxml.etree.parse(self.outputFilenames[1])
            dataContainer = sourceFile.getroot()

            for conceptName in sorted(self.extractedConcepts.keys()):
                # Remove old concept data.
                dataSubContainer = dataContainer.find(dataContainer.tag.rstrip('s') + '[@name="' + conceptName
                                                      + ' (' + self.refName + ' ' + self.refVersion + ')' + '"]')

                if dataSubContainer:
                    dataContainer.remove(dataSubContainer)

                # Add new concept data.
                dataSubContainer = lxml.etree.Element(dataContainer.tag.rstrip('s'))
                dataSubContainer.set('name', conceptName + ' (' + self.refName + ' ' + self.refVersion + ')')
                dataSubContainer.text = '\n'
                dataSubContainer.tail = '\n\n'
                aContainer = lxml.etree.Element('attributes')

                if 'attributes' in self.extractedConcepts[conceptName]:
                    aContainer.text = ','.join(i for i in self.extractedConcepts[conceptName]['attributes'])

                aContainer.tail = '\n'
                dataSubContainer.append(aContainer)
                eContainer = lxml.etree.Element('elements')

                if 'elements' in self.extractedConcepts[conceptName]:
                    eContainer.text = ','.join(i for i in self.extractedConcepts[conceptName]['elements'])

                eContainer.tail = '\n'
                dataSubContainer.append(eContainer)
                tContainer = lxml.etree.Element('types')

                if 'types' in self.extractedConcepts[conceptName]:
                    tContainer.text = ','.join(i for i in self.extractedConcepts[conceptName]['types'])

                tContainer.tail = '\n'
                dataSubContainer.append(tContainer)
                dataContainer.append(dataSubContainer)

            # Write the updated XML structure to the output file.
            outputFile = file(self.outputFilenames[1], 'w')
            outputFile.write(lxml.etree.tostring(lxml.etree.ProcessingInstruction(
                'xml', text='version="1.0" encoding="utf-8"'), encoding='utf-8') + '\n\n')
            outputFile.write(lxml.etree.tostring(sourceFile, encoding='utf-8'))
            outputFile.close()
        except IOError:
            self.outputService.printMessage(message='ERROR: File not found - "' + self.outputFilenames[1] + '"!')
